const AWSLambdaCaller = require('./services/AWSLambdaCaller');
const Util = require('./util');
const MediaResolver = require('./util/MediaResolver');
const DialogFlowAccessor = require('./services/DialogFlowAccessor');
const KeyMatcherAccessor = require('./services/KeyMatcherAccessor');
const RasaAccessor = require('./services/RasaAccessor');
const defaultFallbackIntent = require('./util/defaultFallbackIntent');
const RequestPreprocessor = require('./util/RequestPreprocessor');
const _ = require("lodash");
const md5 = require("md5");
const IntentResolver = require('./util/IntentResolver');
const uuid = require('uuid');

class AsksuiteProcessingNaturalUtils {

    constructor(config) {
        this.config = config;

        if (!_.isEmpty(_.get(this.config, "redis"))) {
            this.redis = this.config.redis;
        }
    }

    withRedis(){
        return !_.isEmpty(_.get(this.config, "redis"));
    }

    async getInCacheDialogflow(request) {
        if (!this.withRedis()) {
            return Promise.resolve(null);
        }

        const hash = this.generateHash(request);

        const data = await this.getInCache(`dialogflow:${hash}`);

        if (!data) {
            return Promise.resolve(null);
        }

        const {intent} = data;
        console.log(`OBTAINED BY CACHE ${intent}`);

        return {intent};
    }

    findDialogByIntent(request, intent){
        const {intents, companyId, text} = request;

        const intentResolver = new IntentResolver(intents, companyId, text);
        return intentResolver.findDialogByIntent(intent);
    }

    generateHash(request) {
        const {text} = request;

        const plainText = _.chain(text)
            .deburr() // special characters
            .toLower() // lowercase
            .value() // get value from chain
            .replace(/\s{2,}/g, " ") // remove duplicate spaces
            .replace(/[^\w\/@ ]/g, ""); // remove all characters dont in (word, /, @, space)

        return md5(plainText);
    }

    async saveInCacheDialogflow(intent, request) {
        if (!this.withRedis()) {
            return Promise.resolve(null);
        }

        if (Util.isNaoentendi(intent)) {
            return Promise.resolve(null);
        }

        const {text} = request;
        const hash = this.generateHash(request);

        return this.setInCache(`dialogflow:${hash}`, {
            intent,
            text,
            hash,
        });
    }

    async getInCache(key) {
        return this.redis.getValue(`${key}`);
    }

    async setInCache(key, intent) {
        return this.redis.setValueWithExpirationTime(`${key}`, intent, true, AsksuiteProcessingNatural.EXPIRATION_TIME);
    }

}

class AsksuiteProcessingNatural {
    static get ORDERED_ACCESSORS() { return ['DIALOG_FLOW', 'RASA', 'KEY_MATCHER']; } // Robot will search for intents in this order default
    static get RASA_FIRST_ORDERED_ACCESSORS() { return ['RASA', 'DIALOG_FLOW', 'KEY_MATCHER']; } // Robot will search for intents in this order if tryRasaFirst option is set to true
    static get EXPIRATION_TIME() { return 60 * 60 * 3 }

    constructor(config) {
        this.config = config;
        this.utils = new AsksuiteProcessingNaturalUtils(this.config);
    };

    findIntent(request) {

        return new  Promise(async (resolve, reject) => {
            let intent = {};

            request = await RequestPreprocessor.process(request);

            const awsLambdaCaller = new AWSLambdaCaller(request.config.CORE_LAMBDA_AWS);
            const dialogFlowAccessor = new DialogFlowAccessor(awsLambdaCaller);
            const keyMatcherAccessor = new KeyMatcherAccessor(awsLambdaCaller);
            const rasaAccessor = new RasaAccessor(request.config.RASA_URL, request.config.RASA_CONFIDENCE_THRESHOLD);

            let orderOption = AsksuiteProcessingNatural.ORDERED_ACCESSORS;
            if (request.tryRasaFirst) {
                orderOption = AsksuiteProcessingNatural.RASA_FIRST_ORDERED_ACCESSORS;
            }

            request.findIntent = defaultFallbackIntent(request.companyId);

            for (const accessorName of orderOption) {
                switch (accessorName) {
                    case 'DIALOG_FLOW':
                        const uuid4 = uuid.v4();
                        console.time(`[${uuid4}] fetch dialog flow`);
                        intent = await this.utils.getInCacheDialogflow(request);
                        if (!intent) {
                            intent = await dialogFlowAccessor.resolveText(request);
                            await this.utils.saveInCacheDialogflow(intent.intent, request);
                        }
                        intent = this.utils.findDialogByIntent(request, intent.intent);
                        console.timeEnd(`[${uuid4}] fetch dialog flow`);
                        break;
                    case 'KEY_MATCHER':
                        intent = await keyMatcherAccessor.resolveText(request);
                        intent.keywordTry = { try: true };
                        break;
                    case 'RASA':
                        intent = await rasaAccessor.resolveText(request);
                        break;
                }

                request.findIntent = intent;

                if (intent.intent && !Util.isNaoentendi(intent.intent)) {
                    intent.resolver = accessorName;

                    if (intent.resolver === 'KEY_MATCHER') {
                        intent.keywordTry.match = intent.intent;
                        intent.keywordTry.intent = intent.intent;
                    }

                    //FOR exit condition
                    break;
                }

            }

            resolve(intent);

        })

    }

    resolveMedia(request) {
        return new Promise (async (resolve, reject) => {
            try {
                const mediaResolver = new MediaResolver(request.companyId);
                const intent = await mediaResolver.resolve(request.media, request.intents);
                resolve(intent)
            } catch (e) {
                reject(e);
            }
        });
    }
}

module.exports = AsksuiteProcessingNatural;
